import numpy as np

from transit_chem.basis import Basis

from . import transittimeanalyzer, utils


class ThreeD(transittimeanalyzer.TransitTime):
    def __init__(
        self, nregions, boundaries, molecule, basis: Basis, id: str = "unlabeled"
    ):
        """
        :param boundaries: A list of the x coordinates of planes dividing the regions.
        :type boundaries: iterable

        """
        super(ThreeD, self).__init__(nregions)  # Call base class __init__
        self.Ndim = 3
        self.molecule = molecule

        #  These are the limits of integration in x for 'all space'. The basis class takes care of the other
        #  dimensions, but has no easy way to do it in x due to the integrals over the various regions.
        #
        self.neg_infty = -3 + min(atom.x for atom in self.molecule)
        self.pos_infty = 3 + max(atom.x for atom in self.molecule)

        bounds = sorted([self.neg_infty, self.pos_infty] + list(boundaries))
        self.boundaries = list(utils.pairwise(bounds))

        self.basis = basis
        self.Nbasis = len(self.basis)
        self.basis_funcs = self.basis
        self.initial_state = None
        self.id = id

    def get_H(self):
        """Create the Hamiltonian under the Extended Hueckel Approximation."""
        self.get_S()
        self.H = np.zeros((self.Nbasis, self.Nbasis))

        for i in range(self.Nbasis):
            self.H[i, i] = -self.molecule.atoms[i].VOIE

        for i in range(self.Nbasis):
            for j in range(self.Nbasis):
                if i != j:
                    self.H[i, j] = (
                        1.75 * (self.H[i, i] + self.H[j, j]) * self.S[i, j] / 2.0
                    )

    def get_S(self):
        """Create the overlap matrix"""
        self.S = np.zeros((self.Nbasis, self.Nbasis))
        for i in range(self.Nbasis):
            for j in range(i + 1):
                self.S[i, j] = self.S[j, i] = self.basis[i].overlap(self.basis[j])

        if not np.all(np.isclose(np.diag(self.S), 1.0, atol=1e-2)):
            print(self.S)
            raise ValueError("Basis is not normalized!")

    def create_mo_func(self, i):
        """Create the i'th MO function, f(x, y, z)."""

        def mo_func(x, y, z):
            val = 0.0
            for j in range(self.Nbasis):
                val += self.basis_funcs[j](x, y, z) * self.eigvecs[j, i]
            return val

        return mo_func

    def get_eigen_basis(self):
        """Create the eigenbasis as a list of functions."""
        self.eigenbasis = []
        for i in range(self.Nbasis):
            mo_func = self.create_mo_func(i)
            self.eigenbasis.append(mo_func)

    def calc_partial_overlap(self):
        """Calculate the partial overlap integrals in the MO basis"""
        self.S_partial = np.zeros((self.Nregions, self.Nbasis, self.Nbasis))
        S_partial_ao = np.zeros((self.Nregions, self.Nbasis, self.Nbasis))

        # Integrate the overlap in the AO basis over each region.
        # S is symmetric so only calculate the upper triangular form and copy.
        for n_region in range(self.Nregions):
            for i in range(self.Nbasis):
                for j in range(i + 1):
                    xmin, xmax = self.boundaries[n_region]
                    Sij = self.basis[i].overlap(self.basis[j], xmin=xmin, xmax=xmax)
                    S_partial_ao[n_region, i, j] = S_partial_ao[n_region, j, i] = Sij

        assert np.allclose(np.sum(S_partial_ao, axis=0), self.S, atol=1e-2)

        # Transform from ao -> mo basis
        for n_region in range(self.Nregions):
            self.S_partial[n_region] = self.matrix_ao_to_mo(S_partial_ao[n_region])

        #  Adding up partial overlaps in MO basis should give the identity matrix.
        #  Since eigenfunctions of hermitian operators form an orthogonal set.
        S = np.sum(self.S_partial, axis=0)
        assert np.allclose(S, np.eye(*S.shape), atol=1e-2)

    def func_to_ao(self, func):
        """Given a vector of values at each grid point, return coefficients in AO basis.

            Currently only works for PzBasis classes.
        """
        ao_expansion_coeffs = np.zeros(self.Nbasis)

        for i, basis in enumerate(self.basis):
            ao_expansion_coeffs[i] = basis.overlap(func)
        ao_expansion_coeffs /= np.linalg.norm(ao_expansion_coeffs)
        return ao_expansion_coeffs

    def func_to_mo(self, func):
        """Given a 3D function f(x, y, z), return its expansion coefficients in MO basis"""
        ao_expansion_coeffs = self.func_to_ao(func)
        mo_expansion_coeffs = self.ao_to_mo(ao_expansion_coeffs)
        return mo_expansion_coeffs

    def set_initial_state(self, initial_state):
        """Set the initial state of the time-dependent calculation.

        :param initial_state: 3D function of cartesian coordinates f(x, y, z)
        :type initial_state: function
        """
        self.initial_state = initial_state
        self.state_mo_coefficients = self.func_to_mo(self.initial_state)



import os

import numpy as np

from . import units


class Atom(object):
    """Atom class containing coordinates, basis and mass."""

    dir = os.path.dirname(__file__)
    fname = os.path.join(dir, "elements.dat")
    with open(fname) as f:
        splitlines = [line.split(" ") for line in f.readlines()]
        Z, symbols = zip(*splitlines)
        Z = [int(i) for i in Z]
        symbols = [s.strip() for s in symbols]

    ZtoSymbol = {z: symbol for z, symbol in zip(Z, symbols)}
    SymboltoZ = {symbol: z for z, symbol in zip(Z, symbols)}
    minimal_pz_orbital_coeff_dict = {
        1: None,
        6: 1.5679,
        7: 1.9170,
        8: 2.2266,
    }

    # Atomic number to Valence Orbital Ionization Energy in eV
    VOIEdict_eV = {
        1: None,
        6: 10.77,
        7: 13.19,
        8: 15.80,
    }

    def __init__(self, atom_type, coords, basistype="minimalpz"):
        """Constructor for Atom class. Currently supports only minimal pz basis

        :param atom_type: Atom symbol or atomic number.
        :type atom_type: str or int
        :param coords: (x, y, z) of atom.
        :type coords: iterable
        :param basistype: String describing the basis. Only 'minimalpz' implemented
        so far.
        :type basistype: string
        """

        self.Z = self.interprete_atom_type(atom_type)
        self.coords = coords
        self.alpha = self.minimal_pz_orbital_coeff_dict[self.Z]
        self.basis_type = basistype
        self.VOIE = self.get_voie()

    @property
    def symbol(self):
        return __class__.ZtoSymbol[self.Z]

    @property
    def x(self):
        return self.coords[0]

    @property
    def y(self):
        return self.coords[1]

    @property
    def z(self):
        return self.coords[2]

    def translate(self, x, y, z):
        self.coords = (self.coords[0] + x, self.coords[1] + y, self.coords[2] + z)

    def scale(self, factor):
        self.coords = tuple(factor * x for x in self.coords)

    def rotate(self, R):
        coords = np.asarray(self.coords)
        coords = R @ coords
        self.coords = tuple(coords)

    def flip_x(self):
        self.coords = (-self.coords[0], self.coords[1], self.coords[2])

    def interprete_atom_type(self, z):
        """Convert atom_type to atomic number. Accepts symbol or number.

        :param z: Atom symbol or atomic number.
        :type z: str or int

        """
        try:
            z = int(z)
        except ValueError:
            try:
                z = self.SymboltoZ[z]
            except KeyError:
                raise ValueError("Atom type not understood.")
        return z

    def get_voie(self):
        """Return the Valence Orbital Ionization Energy in Hartrees."""
        voie = self.VOIEdict_eV[self.Z]
        if voie is not None:
            return units.conversion_factors["EV_TO_HARTREE"] * voie
        else:
            return None

    def __str__(self):
        s = "{} {:10.5f} {:10.5f} {:10.5f}".format(self.symbol, *self.coords)
        return s


def main():
    a = Atom("H", coords=(4, 5, 6))
    print(a)


if __name__ == "__main__":
    main()



from typing import Tuple

import numpy as np

__all__ = [
    "pairwise_array_from_func",
    "parabola_from_points",
]


def pairwise_array_from_func(items, func, *args, symmetric=False, **kwargs):
    """Create a pairwise array by applying a function to all pairs of items.


    Parameters
    -----------
    items
        A container from which pairs will be generated. Must support len() and integer indexing over range(len(items))
    func
        A function f(first, second, *args, **kwargs) which takes 2 items and returns a float.
    symmetric
        Whether the resulting matrix should be symmetric. If true,
        will only compute each (i, j) pair once and set both [i, j] and [j, i] to that value.

    Returns
    --------
    np.array
        The resulting matrix

    Examples
    ---------

    >>> from transit_chem import pairwise_array_from_func
    >>> def distance(i, j):
    ...     return abs(i - j)
    ...
    >>> pairwise_array_from_func([1, 2, 4], distance)
    array([[0., 1., 3.],
           [1., 0., 2.],
           [3., 2., 0.]])
    >>> pairwise_array_from_func([1, 2, 4], distance, symmetric=True)
    array([[0., 1., 3.],
           [1., 0., 2.],
           [3., 2., 0.]])


    """
    n = len(items)
    result = np.zeros((n, n))

    if symmetric:
        for i in range(n):
            for j in range(i + 1):
                result[i, j] = result[j, i] = func(items[i], items[j], *args, **kwargs)
    else:
        for i in range(n):
            for j in range(n):
                result[i, j] = func(items[i], items[j], *args, **kwargs)
    return result


def parabola_from_points(
    point1: Tuple[float, float],
    point2: Tuple[float, float],
    point3: Tuple[float, float],
):
    """Return a coefficients for a parabola passing through 3 points.

    Parameters
    -----------
    point1
        x, y point
    point2
        x, y point
    point3
        x, y point

    Returns
    --------
    np.array
        The coefficients [a, b, c] for the parabola f(x) = ax^2 + bx + c.

    """
    x1, y1 = point1
    x2, y2 = point2
    x3, y3 = point3

    # The polynomial coefficients are the solution to the
    # linear equation Ax = b where
    # x = [a, b, c]
    A = np.array([[x1 ** 2, x1, 1], [x2 ** 2, x2, 1], [x3 ** 2, x3, 1]])
    b = np.array([y1, y2, y3])
    coeffs = np.linalg.solve(A, b)
    return coeffs




import math
from dataclasses import dataclass
from typing import Tuple

import numba
import numpy as np
from scipy import integrate, special
from scipy.misc import derivative

from . import utils
from transit_chem.validation import check_range_inclusive
from transit_chem import config

___all__ = [
    "HarmonicOscillatorWaveFunction",
    "overlap1d",
]


class Basis:
    def __init__(self, factory, x0, y0, z0, alpha):
        """Create a basis object from some parameters and a function-factory function.

        params
        ======
            factory: function g(x0, y0, z0, alpha) that returns a function f(x, y, z)
            x0, y0, z0: centers of basis function
            alpha: parameter for basis (decay coefficient or some other such nonsense).
        """
        self.f = factory(x0, y0, z0, alpha)
        self.x0 = x0
        self.y0 = y0
        self.z0 = z0
        self.alpha = alpha

    def __call__(self, x, y, z):
        return self.f(x, y, z)

    def overlap(
        self,
        other,
        distance_cutoff=5.0,
        interval=5.0,
        xmin=None,
        xmax=None,
        ymin=None,
        ymax=None,
        zmin=None,
        zmax=None,
    ):
        """Return the overlap of this with another

        This is to reduce the calculations performed by naively integrating over
        all space for all combinations of basis functions.
        Return 0.0 if the basis functions are separated by more than distance_cutoff.
        Integrate from interval away from the centers. If bounds are not given.

        If the midpoint of the centers of the functions is more than distance_cutoff from the boundaries,
        0.0 is returned.

        """
        assert isinstance(other, __class__)
        r = np.sqrt(
            (self.x0 - other.x0) ** 2
            + (self.y0 - other.y0) ** 2
            + (self.z0 - other.z0) ** 2
        )
        if r > distance_cutoff:
            return 0.0
        else:
            if xmin is None:
                xmin = min(self.x0, other.x0) - interval
            if ymin is None:
                ymin = min(self.y0, other.y0) - interval
            if zmin is None:
                zmin = min(self.z0, other.z0) - interval
            if xmax is None:
                xmax = max(self.x0, other.x0) + interval
            if ymax is None:
                ymax = max(self.y0, other.y0) + interval
            if zmax is None:
                zmax = max(self.z0, other.z0) + interval

            #  Check to see if the midpoint of the overlapped functions
            # falls outside integration range.
            #  The maximum value between the centers,
            # and if it's sufficiently far from the region of integration
            #  we can return 0.0 and not lose anything.

            midpoint = (
                0.5 * (self.x0 + other.x0),
                0.5 * (self.y0 + other.y0),
                0.5 * (self.z0 + other.z0),
            )
            if midpoint[0] > (xmax + interval) or midpoint[0] < (xmin - interval):
                return 0.0
            if midpoint[1] > (ymax + interval) or midpoint[1] < (ymin - interval):
                return 0.0
            if midpoint[2] > (zmax + interval) or midpoint[2] < (zmin - interval):
                return 0.0

            # Finally actually do the overlap if none of the zero-conditions apply
            ovlp = overlap_factory(self.f, other.f)
            integrator = utils.integrate_region_gen_3D(
                [[xmin, xmax], [ymin, ymax], [zmin, zmax]]
            )
            return integrator(ovlp)


def overlap_factory(f1, f2):
    """Generate a jitted function of f1(x, y, z) * f2(x, y, z). Only works on f1, f2 that are numba jitted."""

    @numba.jit("float64(float64, float64, float64)", nopython=True)
    def overlap(x, y, z):
        return f1(x, y, z) * f2(x, y, z)

    return overlap


def pz_gaussian_orbital_factory(x0, y0, z0, alpha):
    """Return a function for a gaussian pz orbital function centered at (x0, y0, z0) and decay coeff alpha.

    f = N (z - z0) * exp(-(ar)^2)

    The pz orbital is compiled at runtime using numba's jit LLVM compiler. As such, it is fast!"""

    # TODO: FIX NORMALIZATION CONSTANT.
    N = alpha ** 2.5

    @numba.jit("float64(float64, float64, float64)", nopython=True)
    def gaussian_orbital(x, y, z):

        # TODO: MAKE THIS FUNCTION WHATEVER IS DESIRED.
        r = math.sqrt((x - x0) ** 2 + (y - y0) ** 2 + (z - z0) ** 2)
        return N * (z - z0) * math.exp(-((alpha * r) ** 2))

    return gaussian_orbital


def pz_orbital_factory(x0, y0, z0, alpha):
    """Return a function for pz orbital function centered at (x0, y0, z0) and decay coeff alpha.

    The pz orbital is compiled at runtime using numba's jit LLVM compiler. As such, it is fast!"""
    N = alpha ** 2.5 / math.sqrt(math.pi)

    @numba.jit("float64(float64, float64, float64)", nopython=True)
    def pz_orbital(x, y, z):
        r = math.sqrt((x - x0) ** 2 + (y - y0) ** 2 + (z - z0) ** 2)
        return N * (z - z0) * math.exp(-alpha * r)

    return pz_orbital


def overlap1d(
    first: callable,
    second: callable,
    *args,
    lower_limit: float = -np.inf,
    upper_limit: float = np.inf,
) -> float:
    """Compute the overlap integral in 1 dimension over the specified range

    Parameters
    -----------
    first
        The first function
    second
        The second function
    *args
        Extra args to pass to **both** first and second function
    lower_limit
        The lower limit of integration
    upper_limit
        The upper limit of integration

    Returns
    --------
    overlap
        The value of the overlap integral.

    Examples
    ---------

    >>> from transit_chem import overlap
    >>> from numpy import sin, pi
    >>>
    >>> def f1(x):
    ...     return sin(x)
    ...
    >>> def f2(x):
    ...     return sin(2*x)
    ...
    >>> overlap(f1, f2, lower_limit=0, upper_limit=2*pi)
    6.869054119163646e-17
    >>> overlap(f1, f1, lower_limit=0, upper_limit=2*pi)
    3.141592653589793

    """

    def integrand(x, *args_):
        return first(x, *args_) * second(x, *args_)

    return integrate.quad(integrand, a=lower_limit, b=upper_limit, args=args)[0]


def kinetic_integral(
    first: callable,
    second: callable,
    *args,
    lower_limit: float = -np.inf,
    upper_limit: float = np.inf,
) -> float:
    """Return kinetic energy integral of two functions

    Parameters
    -----------
    first
        The first function
    second
        The second function
    *args
        Extra args to pass to **both** first and second function
    lower_limit
        The lower limit of integration
    upper_limit
        The upper limit of integration

    Returns
    --------
    float
        The value of the kinetic energy integral.

    """

    def kinetic_func(x, *args_):
        return -0.5 * derivative(second, x, dx=1e-8, n=2, args=args_)

    return overlap1d(
        first, kinetic_func, *args, lower_limit=lower_limit, upper_limit=upper_limit
    )


def make_potential_integral(potential: callable):
    """

    Parameters
    -----------

    potential
        A function V(x) that takes x and returns the potential energy.
:
    """

    def potential_integral(
        first: callable,
        second: callable,
        *args,
        lower_limit: float = -np.inf,
        upper_limit: float = np.inf,
    ) -> float:
        """Return kinetic energy integral of two functions

        Parameters
        -----------
        first
            The first function
        second
            The second function

        *args
            Extra args to pass to **both** first and second function
        lower_limit
            The lower limit of integration
        upper_limit
            The upper limit of integration

        Returns
        --------
        float
            The value of the kinetic energy integral.

        """

        def potential_func(x, *args_):
            return potential(x) * second(x, *args_)

        return overlap1d(
            first,
            potential_func,
            *args,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
        )

    return potential_integral


@dataclass
class HarmonicOscillator:
    """A 1D quantum harmonic oscillator wave function.

    Parameters
    -----------
    n
        The quantum number
    center
        The center of the potential well
    omega
        The angular frequency of the oscillator
    mass
        The mass of the particle

    Examples
    ---------

    >>> ho = HarmonicOscillator(n=2, center=0.5)
    >>> ho
    HarmonicOscillator(n=2, center=0.5, omega=1, mass=1.0)
    >>> ho(0.5)
    -0.5311259660135984
    >>> ho(1000)
    0.0
    >>> ho(-1000)
    0.0

    """

    n: int
    center: float
    mass: float = 1.0
    omega: float = 1.0

    def __post_init__(self):
        errors = [
            check_range_inclusive("n", self.n, 0, config.HARMONIC_OSCILLATOR_MAX_N),
            check_range_inclusive("mass", self.mass, config.SMALL_NUMBER, config.LARGE_NUMBER),
            check_range_inclusive("omega", self.omega, config.SMALL_NUMBER, config.LARGE_NUMBER)
        ]
        if any(errors):
            raise ValueError(" ".join(errors))

    @staticmethod
    def from_potential_points(
        point1: Tuple[float, float],
        point2: Tuple[float, float],
        point3: Tuple[float, float],
        n: int,
        mass: float = 1.0,
    ):
        """Create a harmonic oscillator wave function from 3 samples of the potential.

        The three points are fit to a parabola (harmonic potential), then the parameters
        for the harmonic oscillator are determined, and the corresponding wave function
        generated and returned.

        Parameters
        -----------
        point1
            A sample point of the potential
        point2
            A sample point of the potential
        point3
            A sample point of the potential
        n
            The quantum number of the resulting wave function
        mass
            The mass of the particle

        Returns
        --------
        HarmonicOscillator
            The harmonic oscillator wave function resulting from the potential at the 3
            points.

        Examples
        ---------

        >>> ho = HarmonicOscillator.from_potential_points(
        ...     point1=(0.5, 1),
        ...     point2=(2.0, 0.5),
        ...     point3=(3.0, 1.5),
        ...     n=0
        ... )
        >>> ho
        HarmonicOscillator(n=0, center=1.5624999999999998, omega=1.0327955589886444, mass=1.0)


        """
        a, b, c = utils.parabola_from_points(point1, point2, point3)
        center = -b / (2 * a)

        # a = m/2 * w**2
        # 2a / m = w**2
        # sqrt(2a/m) = w
        w = np.sqrt(2 * a / mass)
        return HarmonicOscillator(n=n, center=center, omega=w, mass=mass)

    @property
    def N(self):
        return (
            1.0
            / math.sqrt(2 ** self.n * math.factorial(self.n))
            * ((self.mass * self.omega) / math.pi) ** 0.25
        )

    @property
    def _hermite(self):
        return special.hermite(self.n)

    def __call__(self, x):
        y = (np.sqrt(self.mass * self.omega)) * (x - self.center)
        return self.N * np.exp(-(y ** 2) / 2.0) * self._hermite(y)



import numpy as np

from transit_chem.utils import parabola_from_points


def triple_well_potential(center1, barrier12, center2, barrier23, center3):
    """Construct a well potential from the points.

    x                                                         x
    x                                                         x
    x                                                         x
    x                                                        x
    x                                                        x
     x          barrier12                                   x
     x         x         x            barrier23            x
      x       x          x           x         x          x
       center1            x         x           x       x
                           x       x             center3
                            center2

    Returns
    --------
    callable
        A function f(x) that returns the y value of the potential at coordinate x.

    Examples
    ---------

    >>> f = triple_well_potential(
    ...     center1=(0, 0),
    ...     barrier12=(0.5, 1),
    ...     center2=(2, 0.5),
    ...     barrier23=(3, 1.5),
    ...     center3=(4, 1.0),
    ... )
    >>> f(0)
    array(0.)
    >>> f(3)
    array(1.5)
    >>> f([1, 2, 3])
    array([0.56666667, 0.5       , 1.5       ])

    """

    # Wells are in expected left to right ordering.
    if not center1[0] < barrier12[0] < center2[0] < barrier23[0] < center3[0]:
        raise ValueError("Points are not in ascending x-value order.")

    # Wells are below the barriers
    assert center1[1] < barrier12[1]
    assert center2[1] < barrier12[1]
    assert center2[1] < barrier23[1]
    assert center3[1] < barrier23[1]

    def fit_well(center, barrier):
        center_x, center_y = center
        barrier_x, barrier_y = barrier
        # Third point is reflecting barrier about center
        x = -barrier_x + 2 * center_x
        y = barrier_y
        return np.poly1d(parabola_from_points(center, barrier, (x, y)))

    well1 = fit_well(center1, barrier12)
    well2 = np.poly1d(parabola_from_points(barrier12, center2, barrier23))
    well3 = fit_well(center3, barrier23)

    # Pre fetch to avoid repeated lookup
    barrier12_x = barrier12[0]
    barrier23_x = barrier23[0]

    def potential(x):
        if x < barrier12_x:
            return well1(x)
        elif barrier12_x <= x <= barrier23_x:
            return well2(x)
        elif x > barrier23_x:
            return well3(x)

    # vectorize so it can work on numpy arrays
    return np.vectorize(potential)


def centers_barriers_from_params(
    well1_depth,
    well1_halfwidth,
    bridge_length,
    bridge_depth,
    well3_halfwidth,
    well3_depth,
):
    center1 = [0, 0]
    barrier1 = [center1[0] + well1_halfwidth, center1[1] + well1_depth]
    center2 = [barrier1[0] + 0.5 * bridge_length, barrier1[1] - bridge_depth]
    barrier2 = [barrier1[0] + bridge_length, barrier1[1]]
    center3 = [barrier2[0] + well3_halfwidth, barrier2[1] - well3_depth]

    centers = np.asarray([center1, center2, center3])
    barriers = np.asarray([barrier1, barrier2])
    return centers, barriers


#
#     def get_eigen_basis(self):
#         """Create the normalized eigen basis from the AO basis and the eigenvectors."""
#         self.eigenbasis = [np.zeros_like(self.basis_vec[0]) for _ in range(self.Nbasis)]
#         for j in range(self.Nbasis):
#             for i in range(self.Nbasis):
#                 self.eigenbasis[j] += self.eigvecs[i, j] * self.basis_vec[i]
#         self.eigenbasis = [i * self.normalize(i) for i in self.eigenbasis]
#
#     def get_H(self):
#         """Calculate the hamiltonian"""
#         self.V = np.zeros((self.Nbasis, self.Nbasis))  # Potential
#         for i in range(self.Nbasis):
#             for j in range(i + 1):
#                 Vij = self.integrate(self.basis_vec[j] * self.potential_array
#                                     *self.basis_vec[i])
#                 self.V[i, j] = self.V[j, i] = Vij
#
#         self.T = np.zeros((self.Nbasis, self.Nbasis))
#         for i in range(self.Nbasis):
#             for j in range(self.Nbasis):
#                 # < i | T | j > = - 0.5 p*p, apply p in each direction since it is Hermitian
#                 # the finite difference behaves better this way
#                 di = utils.np_finite_diff(self.basis_vec[i], self.coords, 1)
#                 dj = utils.np_finite_diff(self.basis_vec[j], self.coords, 1)
#                 # Duplicate the first value so the length is the same.
#                 # This is good approximation with smooth function and
#                 # dense enough grid.
#                 di = np.insert(di, 0, di[0])
#                 dj = np.insert(dj, 0, dj[0])
#                 # negative sign not here since - 0.5 * p * p = -0.5 * i(dxi)*i(dxj) = 0.5(dxi)(dxj)
#                 self.T[i, j] = (0.5 / self.mass) * self.integrate(di * dj)
#
#         self.H = self.V + self.T
#
#     def get_S(self):
#         """Calculate the overlap matrix."""
#         self.S = np.zeros((self.Nbasis, self.Nbasis))
#         for i in range(self.Nbasis):
#             for j in range(i + 1):
#                 Sij = self.overlap(self.basis_vec[j], self.basis_vec[i])
#                 self.S[i, j] = self.S[j, i] = Sij
#
#     def calc_partial_overlap(self):
#         """Calculate the partial overlap integrals in the MO basis."""
#         self.S_partial = np.zeros((self.Nregions, self.Nbasis, self.Nbasis))
#         for n_region in range(self.Nregions):
#             for i in range(self.Nbasis):
#                 for j in range(i + 1):
#                     Sij = self.region_integrators[n_region](
#                         self.eigenbasis[i] * self.eigenbasis[j]
#                     )
#                     self.S_partial[n_region, i, j] = self.S_partial[n_region, j, i] = Sij
#         return self.S_partial
#
#     def vec_to_ao(self, vec):
#         """Given a vector of values at each grid point, return coefficients in AO basis."""
#         vec *= self.normalize(vec)
#         c = np.zeros(self.Nbasis)
#         for i in range(self.Nbasis):
#             c[i] = np.dot(vec, self.basis_vec[i])
#         c /= np.linalg.norm(c)
#         vec_in_ao_basis = np.zeros(len(self.coords))
#         for i in range(self.Nbasis):
#             vec_in_ao_basis += c[i] * self.basis_vec[i]
#         return c, vec_in_ao_basis
#
#     def vec_to_mo(self, vec):
#         """Given a vector of values at each grid point, return coefficients in MO basis.
#
#         :param vec: 1D Vector of values along the self.coords.
#         :type vec: np.ndarray
#         """
#         vec *= self.normalize(vec)
#         c = np.zeros(self.Nbasis)
#         for i in range(self.Nbasis):
#             c[i] = np.dot(vec, self.eigenbasis[i])
#         c /= np.linalg.norm(c)
#         vec_in_mo_basis = np.zeros(len(self.coords))
#         for i in range(self.Nbasis):
#             vec_in_mo_basis += c[i] * self.eigenbasis[i]
#         return c, vec_in_mo_basis * self.normalize(vec_in_mo_basis)

#
#     def normalized(self, vec):
#         """Return a normalized version of given vector.
#
#         Vec is not modified in place.
#
#         :param vec: 1D Vector of values along the self.coords.
#         :type vec: np.ndarray
#         """
#         return 1.0 / math.sqrt(self.overlap(np.conjugate(vec), vec))
#
#     def time_evolve_vec(self, vec):
#         """Given a vector return the time-dependence of the vector as a complex matrix. Time is an input vector
#         of times. Each row of the matrix is the wave function's position vector at the corresponding
#         point in time in the t vector given."""
#         vec *= self.normalize(vec)
#         c, _ = self.vec_to_mo(vec)
#         td = np.zeros((len(self.times), len(self.coords)), dtype=np.complex128)
#         for i in range(self.Nbasis):
#             time_part = np.exp(-1j * self.eigvals[i] * self.times)
#             td += c[i] * np.outer(time_part, self.eigenbasis[i])
#         return td
#
#     def set_initial_state(self, initialstate):
#         """Set the initial state for the time dependent solution.
#
#         :param initialstate: 1D Vector of values along the self.coords.
#         :type initialstate: np.ndarray
#         """
#         initialstate *= self.normalize(initialstate)
#         self.initial_state = initialstate
#         self.state_mo_coefficients = self.vec_to_mo(initialstate)[0]
#         self.initial_state_energy = np.dot(self.state_mo_coefficients**2, self.eigvals)
#
#     def time_evolve_initialstate(self, state):
#         self.time_evolved_state = self.time_evolve_vec(state)
#         self.time_evolved_prob_density = self.time_evolved_state.conj() * self.time_evolved_state



from os import getenv

HARMONIC_OSCILLATOR_MAX_N = int(getenv("TRANSIT_HARMONIC_OSCILLATOR_MAX_N", 50))
SMALL_NUMBER = float(getenv("TRANSIT_SMALL_NUMBER", 1e-8))
LARGE_NUMBER = float(getenv("TRANSIT_LARGE_NUMBER", 1000))


# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['transit_chem']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'numba>=0.46.0,<0.47.0',
 'numpy>=1.17,<2.0',
 'scipy>=1.3,<2.0']

setup_kwargs = {
    'name': 'transit-chem',
    'version': '0.0.123',
    'description': '',
    'long_description': None,
    'author': 'Evan Curtin',
    'author_email': 'fakeemail@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



def check_range_inclusive(name: str, x, min_val=None, max_val=None) -> str:
    
    if (min_val is None) and (max_val is None):
        raise ValueError("Min and max_val can't both be None")
    elif min_val is None:
        if x > max_val:
            return f"{name} too large, got {x}, expected {name} < {max_val}"
    elif max_val is None:
        if x < min_val:
            return f"{name} too small, got {x}, expected {name} > {min_val}"
    else:
        if x < min_val or x > max_val:
            return (
                f"{name} out of range, got {x}" f"expected {min_val} < {name} < {max_val}"
            )

    return ""
